
#_wb_plain_banner, #_wb_frame_top_banner
{
    display: block !important;
    top: 0px !important;
    left: 0px !important;
    font-family: "Open Sans", "Helvetica Neue", Helvetica, Arial, sans-serif !important;
    width: 100% !important;
    font-size: 18px !important;
    background-color: #444 !important;
    color: white !important;
    text-align: center !important;
    z-index: 2147483643 !important;
    line-height: normal !important;
}

#title_or_url {
  white-space: nowrap;
  overflow: hidden;
  text-overflow: ellipsis;
  padding-right: 4px;
  display: block;
}

#_wb_plain_banner
{
    position: absolute !important;
    padding: 4px !important;
    border: 1px solid !important;
}

#_wb_frame_top_banner
{
    position: absolute !important;
    border: 0px;
    height: 44px !important;
}

#wb_iframe_div
{
    position: absolute;
    width: 100%;
    height: 100%;
    padding: 44px 4px 4px 0px;
    border: none;
    box-sizing: border-box;
    -moz-box-sizing: border-box;
    -webkit-box-sizing: border-box;
    overflow: hidden;
}

.wb_iframe
{
    width: 100%;
    height: 100%;
    border: 2px solid #545454;
    padding: 0px 0px 0px 0px;
    overflow: scroll;
}
