<!DOCTYPE html>
<html>
<head>
    <meta http-equiv="content-type" content="text/html; charset=UTF-8;charset=utf-8">
    <title>Pywb Wayback Machine Home Page</title>
    <link rel="stylesheet" href="{{ static_prefix }}/css/bootstrap.min.css">
    <script src="{{ static_prefix }}/js/jquery-latest.min.js"></script>
    <script src="{{ static_prefix }}/js/bootstrap.min.js"></script>
</head>
<body>
<div class="container">
    <div class="row">
        <h2 class="display-2">Pywb Wayback Machine</h2>
        <p class="lead">This archive contains the following collections:</p>
    </div>
    <div class="row">
        <ul>
            {% for route in routes %}
                <li>
                    <a href="{{ env['pywb.app_prefix'] + '/' + route }}">{{ '/' + route }}</a>
                    {% if all_metadata and all_metadata[route] %}
                        ({{ all_metadata[route].title }})
                    {% endif %}
                </li>
            {% endfor %}
        </ul>
    </div>
</div>
</body>
</html>
